# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml
import io
import requests
import zipfile
import unicodedata
import os
from time import strptime
from typing import Any, cast, Dict, List, Optional, Tuple, Union

from spinn_utilities.typing.json import JsonObject

CITATION_FILE_VERSION_FIELD = "version"
CITATION_FILE_DATE_FIELD = "date-released"
CITATION_AUTHORS_TYPE = "authors"
CITATION_AUTHOR_FIRST_NAME = "given-names"
CITATION_AUTHOR_SURNAME = "family-names"
CITATION_FILE_DESCRIPTION = "title"

ZENODO_RELATION_FIELD = "relation"
ZENODO_NEWER_VERSION_OF = 'isNewVersionOf'
ZENODO_SIBLING_OF = "cites"
ZENODO_METADATA = 'metadata'
ZENODO_PRE_RESERVED_DOI = "prereserve_doi"
ZENODO_DOI_VALUE = "doi"
ZENODO_DEPOSIT_ID = 'id'
ZENODO_METADATA_TITLE = "title"
ZENODO_METATDATA_DESC = "description"
ZENODO_METADATA_CREATORS = "creators"
ZENODO_FILE = "file"
ZENODO_AUTHOR_NAME = "name"

AUTHOR_AFFILIATION = "affiliation"
AUTHOR_ORCID = "orcid"
IDENTIFIER = 'identifier'

# pylint: skip-file


class _ZenodoException(Exception):
    """
    Exception from a call to Zenodo.
    """

    def __init__(
            self, operation: str, expected: int, request: requests.Response):
        """

        :param operation: Name of operation tied
        :param expected: Value that had been expected
        :param request: The errored response received
        """
        super().__init__(
            "don't know what went wrong. got wrong status code when trying "
            f"to {operation}. Got error code {request.status_code} "
            f"(when expecting {expected}) "
            f"with response content {request.content!r}")
        self.request = request
        self.expected = expected


class _Zenodo(object):
    """
    Manages low level access to Zenodo.
    """

    # pragma: no cover
    __slots__ = ("__zenodo_token", )

    _BASE_URI = "https://zenodo.org/api"
    _DEPOSIT_GET_URL = _BASE_URI + "/deposit/depositions"
    _DEPOSIT_PUT_URL = _BASE_URI + "/deposit/depositions/{}/files"
    _PUBLISH_URL = _BASE_URI + "/deposit/depositions/{}/actions/publish"
    _CONTENT_TYPE = "Content-Type"
    _JSON = "application/json"
    _ACCESS_TOKEN = 'access_token'
    _RELATED_IDENTIFIERS = 'related_identifiers'
    _VALID_STATUS_REQUEST_GET = 200
    _VALID_STATUS_REQUEST_POST = 201
    _VALID_STATUS_REQUEST_PUBLISH = 202

    def __init__(self, token: str):
        """

        :param token: Token to pass to https://zenodo.org/api
        """
        self.__zenodo_token = token

    @staticmethod
    def _json(r: requests.Response) -> Optional[JsonObject]:
        try:
            return r.json()
        except Exception:  # pylint: disable=broad-except
            return None

    def get_verify(
            self, related: List[Dict[str, str]]) -> Optional[JsonObject]:
        r = requests.get(
            self._DEPOSIT_GET_URL, timeout=10,
            params={self._ACCESS_TOKEN: self.__zenodo_token,
                    self._RELATED_IDENTIFIERS: str(related)},
            json={}, headers={self._CONTENT_TYPE: self._JSON})
        if r.status_code != self._VALID_STATUS_REQUEST_GET:
            raise _ZenodoException(
                "request a DOI", self._VALID_STATUS_REQUEST_GET, r)
        return self._json(r)

    def post_create(
            self, related: List[Dict[str, str]]) -> Optional[JsonObject]:
        r = requests.post(
            self._DEPOSIT_GET_URL, timeout=10,
            params={self._ACCESS_TOKEN: self.__zenodo_token,
                    self._RELATED_IDENTIFIERS: str(related)},
            json={}, headers={self._CONTENT_TYPE: self._JSON})
        if r.status_code != self._VALID_STATUS_REQUEST_POST:
            raise _ZenodoException(
                "get an empty upload", self._VALID_STATUS_REQUEST_POST, r)
        return self._json(r)

    def post_upload(
            self, deposit_id: str, data: Dict[str, Any],
            files: Dict[str, io.BufferedReader]) -> Optional[JsonObject]:
        r = requests.post(
            self._DEPOSIT_PUT_URL.format(deposit_id), timeout=10,
            params={self._ACCESS_TOKEN: self.__zenodo_token},
            data=data, files=files)
        if r.status_code != self._VALID_STATUS_REQUEST_POST:
            raise _ZenodoException(
                "to put files and data into the preallocated DOI",
                self._VALID_STATUS_REQUEST_POST, r)
        return self._json(r)

    def post_publish(self, deposit_id: str) -> Optional[JsonObject]:
        r = requests.post(
            self._PUBLISH_URL.format(deposit_id), timeout=10,
            params={self._ACCESS_TOKEN: self.__zenodo_token})
        if r.status_code != self._VALID_STATUS_REQUEST_PUBLISH:
            raise _ZenodoException(
                "publish the DOI", self._VALID_STATUS_REQUEST_PUBLISH, r)
        return self._json(r)


class CitationUpdaterAndDoiGenerator(object):
    def __init__(self) -> None:
        self.__zenodo: Optional[_Zenodo] = None

    def update_citation_file_and_create_doi(
            self, citation_file_path: str, doi_title: str, create_doi: bool,
            publish_doi: bool, previous_doi: str, zenodo_access_token: str,
            module_path: str) -> None:
        """
        Take a CITATION.cff file and updates the version and
        date-released fields, and rewrites the ``CITATION.cff`` file.

        :param citation_file_path: File path to the ``CITATION.cff`` file
        :param create_doi:
            Whether to use Zenodo DOI interface to grab a DOI
        :param zenodo_access_token: Access token for Zenodo
        :param publish_doi: Whether to publish the DOI on Zenodo
        :param previous_doi: DOI to append the created DOI to
        :param doi_title: Title for the created DOI
        :param module_path: Path to the module to zip up
        """
        self.__zenodo = _Zenodo(zenodo_access_token)

        # data holders
        deposit_id: Optional[str] = None

        # read in YAML file
        with open(citation_file_path, 'r', encoding="utf-8") as stream:
            yaml_file = yaml.safe_load(stream)

        # if creating a DOI, go and request one
        if create_doi:
            doi_id, deposit_id = self._request_doi(previous_doi)
            yaml_file[IDENTIFIER] = doi_id

        # rewrite citation file with updated fields
        with io.open(citation_file_path, 'w', encoding='utf8') as outfile:
            yaml.dump(yaml_file, outfile, default_flow_style=False,
                      allow_unicode=True)

        # if creating a DOI, finish the request and possibly publish it
        if create_doi:
            assert deposit_id is not None
            self._finish_doi(
                deposit_id, publish_doi, doi_title,
                yaml_file[CITATION_FILE_DESCRIPTION], yaml_file, module_path)

    def _request_doi(self, previous_doi: str) -> Tuple[bytes, Any]:
        """
        Go to Zenodo and requests a DOI.

        :param previous_doi: the previous DOI for this module, if exists
        :return: the DOI id, and deposit id
        """
        # create link to previous version (if applicable)
        related = list()
        related.append({
            ZENODO_RELATION_FIELD: ZENODO_NEWER_VERSION_OF,
            IDENTIFIER: previous_doi})

        # get a request for a DOI
        assert self.__zenodo is not None
        self.__zenodo.get_verify(related)

        # get empty upload
        request_data = self.__zenodo.post_create(related)
        assert request_data is not None

        # get DOI and deposit id
        metadata = cast(Dict[str, Dict[str, str]],
                        request_data[ZENODO_METADATA])
        doi_id = unicodedata.normalize(
            'NFKD',
            (metadata[ZENODO_PRE_RESERVED_DOI]
             [ZENODO_DOI_VALUE])).encode('ascii', 'ignore')
        deposition_id = request_data[ZENODO_DEPOSIT_ID]

        return doi_id, deposition_id

    def _finish_doi(
            self, deposit_id: str, publish_doi: bool, title: str,
            doi_description: str, yaml_file: Dict[str, Any],
            module_path: str) -> None:
        """
        Finishes the DOI on Zenodo.

        :param deposit_id: the deposit id to publish
        :param publish_doi: whether we should publish the DOI
        :param title: the title of this DOI
        :param doi_description: the description for the DOI
        :param yaml_file: the citation file after its been read it
        :param module_path: the path to the module to DOI
        """
        zipped_file = None
        assert self.__zenodo is not None
        try:
            zipped_file = self._zip_up_module(module_path)
            with open(zipped_file, "rb") as zipped_open_file:
                files = {ZENODO_FILE: zipped_open_file}
                data = self._fill_in_data(title, doi_description, yaml_file)
                self.__zenodo.post_upload(deposit_id, data, files)
        finally:
            if zipped_file:
                os.remove(zipped_file)

        # publish DOI
        if publish_doi:
            self.__zenodo.post_publish(deposit_id)

    def _zip_up_module(self, module_path: str) -> str:
        """
        Zip up a module.

        :param module_path: the path to the module to zip up
        :return: the filename to the zip file
        """
        if os.path.isfile('module.zip'):
            os.remove('module.zip')

        avoids = [".git", ".gitignore", ".gitattributes", ".travis.yml",
                  ".github", "model_binaries", "common_model_binaries",
                  ".coveragerc", ".idea"]

        with zipfile.ZipFile(
                'module.zip', 'w', zipfile.ZIP_DEFLATED) as module_zip_file:
            self._zip_walker(module_path, avoids, module_zip_file)
        return 'module.zip'

    @staticmethod
    def _zip_walker(module_path: str, avoids: List[str],
                    module_zip_file: zipfile.ZipFile) -> None:
        """
        Traverse the module and its sub-directories and only adds to the
        files to the zip which are not within a avoid directory that.

        :param module_path: the path to start the search at
        :param avoids: the set of avoids to avoid
        :param module_zip_file: the zip file to put into
        """
        for directory_path, _, files in os.walk(module_path):
            for directory_name in directory_path.split(os.sep):
                if directory_name in avoids:
                    break
            else:
                for potential_zip_file in files:
                    # if safe to zip, zip
                    if potential_zip_file not in avoids:
                        module_zip_file.write(
                            os.path.join(directory_path, potential_zip_file))

    @staticmethod
    def _fill_in_data(doi_title: str, doi_description: str,
                      yaml_file: Dict[str, Any]) -> Dict[str, Any]:
        """
        Add in data to the Zenodo metadata.

        :param doi_title: the title of the DOI
        :param doi_description: the description of the DOI
        :param yaml_file: the citation file once read into the system
        :return: dict containing Zenodo metadata
        """
        # add basic meta data
        metadata: Dict[str, Any] = {
            ZENODO_METADATA_TITLE: doi_title,
            ZENODO_METATDATA_DESC: doi_description,
            ZENODO_METADATA_CREATORS: []
        }

        # get author data from the citation file
        for author in yaml_file[CITATION_AUTHORS_TYPE]:
            author_data = {
                ZENODO_AUTHOR_NAME: (
                    author[CITATION_AUTHOR_SURNAME] + ", " +
                    author[CITATION_AUTHOR_FIRST_NAME])
            }
            if AUTHOR_AFFILIATION in author:
                author_data[AUTHOR_AFFILIATION] = author[AUTHOR_AFFILIATION]
            if AUTHOR_ORCID in author:
                author_data[AUTHOR_ORCID] = author[AUTHOR_ORCID]
            metadata[ZENODO_METADATA_CREATORS].append(author_data)
        return {ZENODO_METADATA: metadata}

    @staticmethod
    def convert_text_date_to_date(
            version_month: Union[int, str], version_year: Union[int, str],
            version_day: Union[int, str]) -> str:
        """
        Convert the 3 components of a date into a CFF date.

        :param version_month: version month, in text form
        :param version_year: version year
        :param version_day: version day of month
        :return: the string representation for the CFF file
        """
        return "{}-{}-{}".format(
            version_year,
            CitationUpdaterAndDoiGenerator.convert_month_name_to_number(
                version_month),
            version_day)

    @staticmethod
    def convert_month_name_to_number(version_month: Union[int, str]) -> int:
        """
        Convert a python month in text form to a number form.

        :param version_month: the text form of the month
        :return: the month int value
        :raises ValueError: when the month name is not recognised
        """
        if isinstance(version_month, int):
            return version_month
        elif isinstance(version_month, str):
            try:
                return int(version_month)
            except ValueError as original:
                try:
                    return strptime(version_month, "%B").tm_mon
                except ValueError:
                    try:
                        return strptime(version_month, "%b").tm_mon
                    except ValueError:  # pragma: no cover
                        raise ValueError(
                            f"Value {version_month} not recognised"
                            " as a month") from original
        else:  # pragma: no cover
            raise ValueError(
                f"Value {version_month} not recognised as a month")



# Copyright (c) 2025 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import os
import sys
from typing import cast, Set

import docstring_parser

ERROR_NONE = 0
ERROR_OTHER = ERROR_NONE + 1
ERROR_FILE = ERROR_OTHER + 1


class DocsChecker(object):
    """
    A utility class to check the docs strings for our rules.

    As we use type annotations doc strings should not have types.
    At best they are the same otherwise they are wrong.

    We check that all documented params are actually ones used.
    The inverse is not checked as a good param name is often enough

    """

    __slots__ = [
        "__check_init", "__check_short", "__check_params",
        "__check_properties", "__check_returns",
        "__check_types_in_docs",
        "__error_level", "__file_path", "__file_errors",
        "__functions_errors"]

    def __init__(
            self, *, check_init: bool = True, check_short: bool = True,
            check_params: bool = True, check_properties: bool = True,
            check_returns: bool = True,
            check_types_in_docs: bool = True) -> None:
        """
        Sets up the doc checker.

        Which functions need to be documented is left to pylint to check.
        Currently, that is public methods and public methods of public classes.
        pylint does not insist init methods are documented.

        :param check_init: flag to trigger checking of __init__ methods.
            If True all init methods must have all params documented
            Descriptions not allowed on __init__ files
            as they should be on the class only.
        :param check_short: Flag to trigger checking of a description.
            For public (None init) methods (except setters) with no return
            there must be a short description
        :param check_params: Flag to trigger checking of params.
            If any param is listed all must be.
            The param does not need to be documented.
        :param check_properties: Flag to trigger checking of Properties
            They must include a description.
            They should not have a return annotation
        :param check_returns: Flag to trigger checking of return annotations
            when not a property
        :param check_types_in_docs:
            Flag to trigger checking that doc-string have no types
        """
        self.__error_level = ERROR_NONE
        self.__check_types_in_docs = check_types_in_docs
        self.__check_init = check_init
        self.__check_params = check_params
        self.__check_properties = check_properties
        self.__check_short = check_short
        self.__check_returns = check_returns
        self.__file_path = "None"
        self.__file_errors = 0
        self.__functions_errors = 0

    def check_dir(self, dir_path: str) -> None:
        """
        Checks all py files in this directory including subdirectories.
        """
        for root, dir_names, files in os.walk(dir_path):
            dir_names.sort()
            files.sort()
            for file_name in files:
                if file_name.endswith(".py"):
                    self.check_file(os.path.join(root, file_name))

    def check_file(self, file_path: str) -> None:
        """
        Check the documentation in this file.
        """
        if self.__error_level > ERROR_OTHER:
            self.__error_level = ERROR_OTHER
        self.__file_path = file_path
        with open(file_path, "r", encoding="utf-8") as file:
            raw_tree = file.read()
        try:
            ast_tree = ast.parse(raw_tree, type_comments=True)
        except SyntaxError as ex:
            raise SyntaxError(f"{ex.msg} of {file_path}") from ex
        for node in ast.walk(ast_tree):
            if isinstance(node, ast.FunctionDef):
                self.check_function(node)

    def _check_function(self, node: ast.FunctionDef) -> str:
        """
        Check the documentation in this function.
        """
        if self.__error_level > ERROR_FILE:
            self.__error_level = ERROR_FILE
        _docs = ast.get_docstring(node)
        if _docs is None:
            # pylint does not require init to have docs
            if node.name == "__init__" and self.__check_init:
                param_names = self.get_param_names(node)
                if (len(param_names) > 0 and self.is_not_overload(node) and
                        not self._test_path()):
                    return "missing docstring"
            return ""
        else:
            docs = cast(str, _docs)

        docstring = docstring_parser.parse(docs)
        param_names = self.get_param_names(node)
        error = self._check_params_correct(param_names, docstring)

        # TODO remove when all repositories fixed
        #  if not self.has_returns(node) and len(docstring.many_returns) > 0:
        #    error += "Unexpected returns"

        if (node.name.startswith("_") or self._test_path() or
                self._overrides(node)):
            # these are not included by readthedocs so less important
            return error

        if node.name == "__init__":
            if self.__check_init:
                if len(docstring.params) != len(param_names):
                    error += "Missing params"
                if docstring.short_description is not None:
                    error += "Short description provided."
                if len(docstring.many_returns) > 0:
                    error += ("Unexpected returns")
        elif self.has_returns(node):
            if self.is_property(node):
                if self.__check_properties:
                    if docstring.short_description is None:
                        # docstring_parser can not handle  these
                        # Read The Docs can
                        if not docs.startswith(":math:"):
                            error += "No short description provided."
                    if len(docstring.many_returns) > 0:
                        error += "Unexpected returns"
                        if self.__check_params:
                            error += self._check_params_all_or_none(
                                param_names, docstring)
            else:
                if self.__check_returns:
                    if len(docstring.many_returns) == 0:
                        error += "No returns"
                if self.__check_params:
                    error += self._check_params_all_or_none(
                        param_names, docstring)
        else:
            if self.__check_short:
                if docstring.short_description is None:
                    error += "No short description provided."
            if self.__check_params:
                error += self._check_params_all_or_none(
                    param_names, docstring)

        error += self._check_blank_lines(docs)

        return error

    def check_function(self, node: ast.FunctionDef) -> None:
        """
        Check the documentation in this function.
        """
        if self.__error_level > ERROR_FILE:
            self.__error_level = ERROR_FILE
        error = self._check_function(node)

        if error:
            if self.__error_level < ERROR_FILE:
                print(f"{self.__file_path}")
                self.__file_errors += 1
            self.__error_level = ERROR_FILE
            print(f"\t{node.name} {node.lineno}")
            print(f"\t{error}")
            self.__functions_errors += 1

    def is_property(self, node: ast.FunctionDef) -> bool:
        """
        :return: True if and only if there is a @property decorator
        """
        for decorator in node.decorator_list:
            if isinstance(decorator, ast.Name) and decorator.id == "property":
                return True
        return False

    def is_not_overload(self, node: ast.FunctionDef) -> bool:
        """
        :return: True if and only if there is NO @overload decorator
        """
        for decorator in node.decorator_list:
            if isinstance(decorator, ast.Name) and decorator.id == "overload":
                return False
        return True

    def has_returns(self, node: ast.FunctionDef) -> bool:
        """
        :return: True if and only if there is a return declared
        """
        returns = node.returns
        if isinstance(returns, ast.Constant):
            if returns.value is None:
                return False
        elif isinstance(returns, ast.Name):
            if returns.id == "Never":
                return False
        return True

    def _overrides(self, node: ast.FunctionDef) -> bool:
        """
        Detects the overrides annotation and the extend_docs is not False
        """
        for decorator in node.decorator_list:
            try:
                if (isinstance(decorator, ast.Call) and
                        isinstance(decorator.func, ast.Name) and
                        decorator.func.id == "overrides"):
                    for keyword in decorator.keywords:
                        if keyword.arg == "extend_doc":
                            value = cast(ast.Constant, keyword.value)
                            return bool(value.value)
                    return True
            except AttributeError:
                print(decorator)
        return False

    def _test_path(self) -> bool:
        """
        :returns: True if the path is likely for tests
        """
        test_paths = ["fec_integration_tests", "pacman_test_objects",
                      "spynnaker_integration_tests", "tests", "unittests"]
        for test_path in test_paths:
            check = os.sep + test_path + os.sep
            if check in self.__file_path:
                return True
        return False

    def _check_params_correct(
            self, param_names: Set[str],
            docstring: docstring_parser.common.Docstring) -> str:
        """
        Checks that all params listed are used and not typed
        """
        error = ""
        for param in docstring.params:
            if param.arg_name not in param_names:
                error += f"{param.arg_name}: is incorrect "
            elif param.type_name:
                if self.__check_types_in_docs:
                    error += f"{param.arg_name}: is typed "
        return error

    def _check_params_all_or_none(
            self, param_names: Set[str],
            docstring: docstring_parser.common.Docstring) -> str:
        """
        Checks that either no or all params are listed
        """
        if len(docstring.params) == 0:
            return ""
        elif len(docstring.params) != len(param_names):
            return "Missing params"
        else:
            return ""

    def _check_blank_lines(self, docs: str) -> str:
        """
        Check there are blank lines where needed
        """
        error = ""

        if self.__check_types_in_docs:
            for key in [":type", ":rtype"]:
                if key in docs:
                    error += f"found {key} "
        index = sys.maxsize
        for key in [":param", ":return", ":raises"]:
            if key in docs:
                key_index = docs.index(key)
                if key_index < index:
                    index = key_index
        if index < sys.maxsize:
            while index > 1 and docs[index-1] in [" ", "\t"]:
                index -= 1
            if index >= 2:
                if docs[index-2: index] != "\n\n":
                    error += "Missing blank line after description"

        return error

    def get_param_names(self, node: ast.FunctionDef) -> Set[str]:
        """
        Gets the names of the parameters found in the abstract syntax tree.

        These are the ones actually declared.

        :returns: Names of all parameter including normal and kwargs ones
        """
        param_names: Set[str] = set()
        for arg in node.args.args:
            param_names.add(arg.arg)
        for arg in node.args.kwonlyargs:
            param_names.add(arg.arg)
        if node.args.vararg:
            param_names.add(node.args.vararg.arg)
        if node.args.kwarg:
            param_names.add(node.args.kwarg.arg)

        if "self" in param_names:
            param_names.remove("self")
        if "cls" in param_names:
            param_names.remove("cls")
        return param_names

    def check_no_errors(self) -> None:
        """
        Checks that there are no errors found.

        Does not run any checks just check status after they are run

        :raises AssertionError: If any previous check found an error
        """
        if self.__error_level > ERROR_NONE:
            raise AssertionError(
                f"The documentation checker found "
                f"{self.__functions_errors} errors "
                f"in {self.__file_errors} files")


if __name__ == "__main__":
    checker = DocsChecker(
        check_init=False,
        check_short=False,
        check_params=False,
        check_properties=False
    )
    # checker.check_dir("")
    checker.check_file("")



# Copyright (c) 2025 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import os
from typing import Collection, Dict, List, Set

from spinn_utilities.config_holder import get_default_cfgs
from spinn_utilities.configs.camel_case_config_parser import (
    CamelCaseConfigParser, TypedConfigParser)
from spinn_utilities.exceptions import ConfigException


class ConfigChecker(object):
    """
    Checks the py and cfg files after the defaults have been set,
    """

    __slots__ = ["_configs",
                 "_default_cfgs",
                 "_directories",
                 "_file_path",
                 "_lines",
                 "_used_cfgs"]

    def __init__(self, directories: Collection[str]) -> None:
        """
        :param directories: Path to find the cfg and py files to check
        """
        # SpiNNGym needs check_all_used
        self._configs = TypedConfigParser()
        self._default_cfgs = get_default_cfgs()
        for default_cfg in self._default_cfgs:
            self._configs.read(default_cfg)
        self._directories = directories
        self._file_path = ""
        self._lines: List[str] = []
        self._used_cfgs: Dict[str, Set[str]] = defaultdict(set)

    def check(self, local_defaults: bool = True) -> None:
        """
        Runs the checks of py and cfg files

        :param local_defaults:
        """
        if local_defaults:
            self._check_defaults()
        self._read_files()
        if local_defaults:
            self._check_all_used()

    def _check_defaults(self) -> None:
        """
        Testing function to identify any configuration options /
        in multiple default files.

        :raises ConfigException:
            If two defaults configuration files set the same value
        """
        config1 = CamelCaseConfigParser()
        for default in self._default_cfgs[:-1]:
            config1.read(default)
        config2 = CamelCaseConfigParser()
        config2.read(self._default_cfgs[-1])
        for section in config2.sections():
            for option in config2.options(section):
                if config1.has_option(section, option):
                    raise ConfigException(
                        f"cfg:{self._default_cfgs} "
                        f"repeats [{section}]{option}")

    def _read_files(self) -> None:
        for directory in self._directories:
            for root, _, files in os.walk(directory):
                for file_name in files:
                    if file_name.endswith(".cfg"):
                        self._file_path = os.path.join(root, file_name)
                        self._check_cfg_file()
                    elif file_name.endswith(".py"):
                        self._file_path = os.path.join(root, file_name)
                        self._check_python_file()

    def _check_cfg_file(self) -> None:
        """
        Support method for :py:func:`check_cfgs`.

        :raises ConfigException: If an unexpected option is found
        """
        if self._file_path in self._default_cfgs:
            return

        config2 = TypedConfigParser()
        config2.read(self._file_path)
        for section in config2.sections():
            if not self._configs.has_section(section):
                raise ConfigException(f"cfg:{self._file_path} has "
                                      f"unexpected section [{section}]")
            for option in config2.options(section):
                if not self._configs.has_option(section, option):
                    raise ConfigException(
                        f"cfg:{self._file_path} "
                        f"has unexpected options [{section}]{option}")

    def _check_python_file(self) -> None:
        """
        A testing function to check that all the `get_config` calls work.

        :raises ConfigException:
            If an unexpected or uncovered `get_config` found
        """
        with open(self._file_path, 'r', encoding="utf-8") as py_file:
            self._lines = list(py_file)
            for index, line in enumerate(self._lines):
                if ("skip_if_cfg" in line):
                    parts = self._get_parts(line, index, "skip_if_cfg")
                    self._check_lines(parts)
                elif ("configuration.get" in line):
                    parts = self._get_parts(line, index, "configuration.get")
                    self._check_lines(parts)
                if ("get_report_path(" in line):
                    parts = self._get_parts(line, index, "get_report")
                    self._check_get_report_path(parts)
                if ("get_timestamp_path(" in line):
                    parts = self._get_parts(line, index, "get_timestamp")
                    self._check_get_report_path(parts)
                if "get_config" not in line:
                    continue
                if (("get_config_bool(" in line) or
                        ("get_config_bool_or_none(" in line) or
                        ("get_config_float(" in line) or
                        ("get_config_float_or_none(" in line) or
                        ("get_config_int(" in line) or
                        ("get_config_int_or_none(" in line) or
                        ("get_config_str(" in line) or
                        ("get_config_str_or_none(" in line) or
                        ("get_config_str_list(" in line)):
                    parts = self._get_parts(line, index, "get_config")
                    self._check_lines(parts)

    def _get_parts(self, line: str, index: int, start: str) -> List[str]:
        while ")" not in line:
            index += 1
            line += self._lines[index]
        parts = line[line.find("(", line.find(start)) + 1:
                     line.find(")")].split(",")
        return parts

    def _check_lines(self, parts: List[str]) -> None:
        section = parts[0].strip().replace("'", "").replace('"', '')
        for i in range(1, len(parts)):
            option = parts[i].strip()
            if option[0] == "'":
                option = option.replace("'", "")
            elif option[0] == '"':
                option = option.replace('"', '')
            else:
                return

            if option.startswith("@"):
                raise ConfigException(
                    f"{self._file_path} has {option=} which starts with @")
            if not self._configs.has_option(section, option):
                raise ConfigException(
                    f"{self._file_path} has unexpected {section=} {option=}")

            self._used_cfgs[section].add(option)

    def _check_get_report_path(self, parts: List[str]) -> None:
        section = "Reports"
        option = "No Option found"
        for part in parts:
            part = part.strip()
            if "=" not in part:
                option = part
            elif part.startswith("option="):
                option = part[7:]
            elif part.startswith("section="):
                section = part[8:]
            elif part.startswith("is_dir="):
                pass
            elif part.startswith("n_run="):
                pass
            else:
                if self._file_path.endswith("config_holder.py"):
                    return
                raise ConfigException(
                    f"in {self._file_path} unexpected {parts=}")

        if option == "option":
            return

        option = option.replace("'", "").replace('"', '')
        section = section.replace("'", "").replace('"', '')

        if not self._configs.has_option(section, option):
            if self._file_path.endswith("config_checker.py"):
                return
            raise ConfigException(
                f"{self._file_path} has unexpected {section=} {option=}")

        self._used_cfgs[section].add(option)

    def _check_all_used(self) -> None:
        current_config = TypedConfigParser()
        current_config.read(self._default_cfgs[-1])
        for section in current_config:
            if section not in self._used_cfgs:
                if section == "DEFAULT":
                    continue
                raise ConfigException(
                    f"in {self._default_cfgs[-1]} {section=} was never used")
            for option in current_config.options(section):
                if option.startswith("@"):
                    continue
                if option not in self._used_cfgs[section]:
                    raise ConfigException(
                        f"in {self._default_cfgs[-1]} cfg "
                        f"{section=} {option=} was never used")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from configparser import NoOptionError
import logging
import os
from typing import List, Optional, Tuple

import appdirs

import spinn_utilities.conf_loader as conf_loader
from spinn_utilities.data import UtilsDataView
from spinn_utilities.configs import CamelCaseConfigParser
from spinn_utilities.configs.no_config_found_exception import (
    NoConfigFoundException)
from spinn_utilities.configs.two_user_configs_exception import (
    TwoUserConfigsException)
from spinn_utilities.exceptions import ConfigException
from spinn_utilities.log import (
    FormatAdapter, ConfiguredFilter, ConfiguredFormatter)

logger = FormatAdapter(logging.getLogger(__file__))

# pylint: disable=global-statement
# Any cleaner method than global statements would add extra overhead
__config: Optional[CamelCaseConfigParser] = None
__default_config_files: List[str] = []
__config_file: Optional[str] = None
__template: Optional[str] = None
__unittest_mode: bool = False


def add_default_cfg(default: str) -> None:
    """
    Adds an extra default configuration file to be read after earlier ones.

    :param default: Absolute path to the configuration file
    """
    if default not in __default_config_files:
        __default_config_files.append(default)


def add_template(template: str) -> None:
    """
    Adds an extra default configuration file to be read after earlier ones.

    :param template: Absolute path to the template file
    """
    global __template
    if __template is None:
        __template = template
    else:
        raise ConfigException("Second template")


def get_default_cfgs() -> Tuple[str, ...]:
    """
    The default configuration files

    This is a read only values to be used outside of normal operations

    :returns: The default configuration files as a tuple.
    """
    return tuple(__default_config_files)


def clear_cfg_files(unittest_mode: bool) -> None:
    """
    Clears any previous set configurations and configuration files.

    After this method :py:func:`add_default_cfg` and :py:func:`set_cfg_files`
    need to be called.

    :param unittest_mode: Flag to put the holder into unit testing mode
    """
    global __config, __config_file, __template, __unittest_mode
    __config = None
    __default_config_files.clear()
    __config_file = None
    __template = None
    __unittest_mode = unittest_mode


def set_cfg_files(config_file: Optional[str], default: str) -> None:
    """
    Adds the configuration files to be loaded.

    :param config_file:
        The base name of the configuration file(s).
        Should not include any path components.
        Use None to not read any file
    :param default:
        Full path to the extra file to get default configurations from.
    """
    global __config_file
    __config_file = config_file
    add_default_cfg(default)


def _pre_load_config() -> CamelCaseConfigParser:
    """
    Loads configurations due to early access to a configuration value.

    :raises ConfigException: Raise if called before setup
    """
    # If you get this error during a unit test, unittest_step was not called
    if not __unittest_mode:
        raise ConfigException(
            "Accessing config values before setup is not supported")
    return load_config()


def logging_parser(config: CamelCaseConfigParser) -> None:
    """
    Create the root logger with the given level.

    Create filters based on logging levels
    """
    try:
        if (has_config_option("Logging", "instantiate") and
                get_config_bool("Logging", "instantiate")):
            level = "INFO"
            if has_config_option("Logging", "default"):
                level = get_config_str("Logging", "default").upper()
            logging.basicConfig(level=level)
        for handler in logging.root.handlers:
            handler.addFilter(
                ConfiguredFilter(config))  # type: ignore[arg-type]
            handler.setFormatter(ConfiguredFormatter(config))
    except NoOptionError:
        pass


def _user_cfg() -> Optional[str]:
    """
    Defines the list of places we can get configuration files from.

    :return: existing fully-qualified filename
    """
    if __config_file is None:
        return None

    dotname = "." + __config_file
    found = None

    for check in [os.path.join(appdirs.site_config_dir(), dotname),
                  os.path.join(appdirs.user_config_dir(), dotname),
                  os.path.join(os.path.expanduser("~"), dotname)]:
        if os.path.isfile(check):
            if found:
                raise TwoUserConfigsException(
                    f"Two user cfg files found {check} and {found}")
            else:
                found = check
    return found


def _install_cfg_and_error() -> NoConfigFoundException:
    """
    Installs a local configuration file based on the templates and raises
    an exception.

    This method is called when no user configuration file is found.

    It will create a file in the users home directory based on the defaults.
    Then it prints a helpful message and throws an error with the same message.

    :return: Exception to be raised by caller
    """
    assert __config_file is not None
    assert __template is not None
    home_cfg = os.path.join(os.path.expanduser("~"), f".{__config_file}")

    with open(home_cfg, "w", encoding="utf-8") as dst:
        with open(__template, "r", encoding="utf-8") as src:
            dst.write(src.read())
            dst.write("\n")
        dst.write("\n# Additional config options can be found in:\n")
        for source in __default_config_files:
            dst.write(f"# {source}\n")
        dst.write("\n# Copy any additional settings you want to change"
                  " here including section headings\n")

    msg = ('Unable to find config file in your home directory\n'
           '**********************************************************\n'
           f'{home_cfg} has been created. \n'
           'Please edit this file and change "machineName" '
           'to the IP address of your SpiNNaker board '
           "or change spalloc_server to the server urls "
           'and change "version" to the version of '
           'SpiNNaker hardware you are running on:\n'
           '***********************************************************\n')
    print(msg)
    return NoConfigFoundException(msg)


def load_config() -> CamelCaseConfigParser:
    """
    Reads in all the configuration files, resetting all values.

    :raises ConfigException: If called before setting defaults
    :returns: A fully loaded parser object
    """
    global __config
    if not __default_config_files:
        raise ConfigException("No default configs set")

    if __config_file:
        user_cfg = _user_cfg()
        if not user_cfg:
            if __template:
                raise _install_cfg_and_error()
            elif __config_file:
                logger.info(f"No default configs {__config_file} "
                            f"found in home directory")
        __config = conf_loader.load_config(
            __config_file, user_cfg, __default_config_files)
    else:
        __config = conf_loader.load_defaults(__default_config_files)

    logging_parser(__config)
    return __config


def is_config_none(section: str, option: str) -> bool:
    """
    Check if the value of a configuration option would be considered None

    :param section: What section to get the option from.
    :param option: What option to read.
    :return: True if and only if the value would be considered None
    """
    value = get_config_str_or_none(section, option)
    return value is None


def get_config_str(section: str, option: str) -> str:
    """
    Get the string value of a configuration option.

    :param section: What section to get the option from.
    :param option: What option to read.
    :return: The option value
    :raises ConfigException: if the Value would be None
    """
    value = get_config_str_or_none(section, option)
    if value is None:
        raise ConfigException(f"Unexpected None for {section=} {option=}")
    return value


def get_config_str_or_none(section: str, option: str) -> Optional[str]:
    """
    Get the string value of a configuration option.

    :param section: What section to get the option from.
    :param option: What option to read.
    :return: The option value
    :raises ConfigException: if the Value would be None
    """
    if __config is None:
        return _pre_load_config().get_str(section, option)
    else:
        return __config.get_str(section, option)


def get_config_str_list(
        section: str, option: str, token: str = ",") -> List[str]:
    """
    Get the string value of a configuration option split into a list.

    :param section: What section to get the option from.
    :param option: What option to read.
    :param token: The token to split the string into a list
    :return: The list (possibly empty) of the option values
    """
    if __config is None:
        return _pre_load_config().get_str_list(section, option, token)
    else:
        return __config.get_str_list(section, option, token)


def get_config_int(section: str, option: str) -> int:
    """
    Get the integer value of a configuration option.

    :param section: What section to get the option from.
    :param option: What option to read.
    :return: The option value
    :raises ConfigException: if the Value would be None
    """
    value = get_config_int_or_none(section, option)
    if value is None:
        raise ConfigException(f"Unexpected None for {section=} {option=}")
    return value


def get_config_int_or_none(section: str, option: str) -> Optional[int]:
    """
    Get the integer value of a configuration option.

    :param section: What section to get the option from.
    :param option: What option to read.
    :return: The option value
    :raises ConfigException: if the Value would be None
    """
    if __config is None:
        return _pre_load_config().get_int(section, option)
    else:
        return __config.get_int(section, option)


def get_config_float(section: str, option: str) -> float:
    """
    Get the float value of a configuration option.

    :param section: What section to get the option from.
    :param option: What option to read.
    :return: The option value.
    :raises ConfigException: if the Value would be None
    """
    value = get_config_float_or_none(section, option)
    if value is None:
        raise ConfigException(f"Unexpected None for {section=} {option=}")
    return value


def get_config_float_or_none(section: str, option: str) -> Optional[float]:
    """
    Get the float value of a configuration option.

    :param section: What section to get the option from.
    :param option: What option to read.
    :return: The option value.
    """
    if __config is None:
        return _pre_load_config().get_float(section, option)
    else:
        return __config.get_float(section, option)


def get_config_bool(section: str, option: str) -> bool:
    """
    Get the Boolean value of a configuration option.

    :param section: What section to get the option from.
    :param option: What option to read.
    :return: The option value.
    :raises ConfigException: if the Value would be None
    """
    value = get_config_bool_or_none(section, option)
    if value is None:
        raise ConfigException(f"Unexpected None for {section=} {option=}")
    return value


def get_config_bool_or_none(section: str, option: str,
                            special_nones: Optional[List[str]] = None
                            ) -> Optional[bool]:
    """
    Get the Boolean value of a configuration option.

    :param section: What section to get the option from.
    :param option: What option to read.
    :param special_nones: What special values to except as None
    :return: The option value.
    :raises ConfigException: if the Value would be None
    """
    if __config is None:
        return _pre_load_config().get_bool(section, option, special_nones)
    else:
        return __config.get_bool(section, option, special_nones)


def set_config(section: str, option: str, value: Optional[str]) -> None:
    """
    Sets the value of a configuration option.

    This method should only be called by the simulator or by unit tests.

    :param section: What section to set the option in.
    :param option: What option to set.
    :param value: Value to set option to
    :raises ConfigException: If called unexpectedly
    """
    if __config is None:
        _pre_load_config().set(section, option, value)
    else:
        __config.set(section, option, value)


def config_sections() -> List[str]:
    """
    :returns: A list of section names
    """
    if __config is None:
        raise ConfigException("configuration not loaded")
    return __config.sections()


def configs_loaded() -> bool:
    """
    :returns: True if and only if the configuration was loaded
    """
    if __config is None:
        return False
    else:
        return True


def has_config_option(section: str, option: str) -> bool:
    """
    Check if the section has this configuration option.

    :param section: What section to check
    :param option: What option to check.
    :return: True if and only if the option is defined. It may be `None`
    """
    if __config is None:
        raise ConfigException("configuration not loaded")
    else:
        return __config.has_option(section, option)


def config_options(section: str) -> List[str]:
    """
    :param section: What section to list options for.
    :returns: a list of option names for the given section name.
    """
    if __config is None:
        raise ConfigException("configuration not loaded")
    return __config.options(section)


def get_report_path(
        option: str, section: str = "Reports", n_run: Optional[int] = None,
        is_dir: bool = False) -> str:
    """
    Gets and fixes the path for this option

    If the cfg path is relative it will be joined with the run_dir_path.

    Creates the path's directory if it does not exist.

    (n_run) and (reset_str) will be replaced

    Later updates may replace other bracketed expressions as needed
    so avoid using brackets in file names

    :param option: cfg option name
    :param section: cfg section. Needed if not Reports
    :param n_run: If provided will be used instead of the current run number
    :param is_dir:
        When true will make sure this path exists as a directory.
        When False will make sure the parent directory is exists.
    :return: An unchecked absolute path to the file or directory
    """
    path = get_config_str(section, option)
    if path.startswith("(global)"):
        path = os.path.join(UtilsDataView.get_global_reports_dir(), path[8:])

    if n_run is not None and n_run > 1:
        if "(n_run)" not in path:
            logger.warning(
                f"cfg option {option} does not have a (n_run) so "
                f"files from different runs may be overwritten")
    if "(n_run)" in path:
        if n_run is None:
            n_run = UtilsDataView.get_run_number()
        path = path.replace("(n_run)", str(n_run))

    if "(reset_str)" in path:
        reset_str = UtilsDataView.get_reset_str()
        path = path.replace("(reset_str)", str(reset_str))

    if "\\" in path:
        path = path.replace("\\", os.sep)

    if not os.path.isabs(path):
        path = os.path.join(UtilsDataView.get_run_dir_path(), path)

    if is_dir:
        os.makedirs(path, exist_ok=True)
    else:
        folder, _ = os.path.split(path)
        os.makedirs(folder, exist_ok=True)

    return path


def get_timestamp_path(option: str, section: str = "Reports") -> str:
    """
    Gets and fixes the path for this option

    If the cfg path is relative it will be joined with the timestamp_path.

    Creates the path's directory if it does not exist.

    Later updates may replace bracketed expressions as needed
    so avoid using brackets in file names

    :param option: cfg option name
    :param section: cfg section. Needed if not Reports
    :return: An unchecked absolute path to the file or directory
    """

    path = get_config_str(section, option)
    if path.startswith("(global)"):
        path = os.path.join(UtilsDataView.get_global_reports_dir(), path[8:])
    elif not os.path.isabs(path):
        path = os.path.join(UtilsDataView.get_timestamp_dir_path(), path)

    folder, _ = os.path.split(path)
    os.makedirs(folder, exist_ok=True)

    return path



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import configparser
from datetime import datetime
import logging
import re
import sys
from typing import (Any, Collection, Dict, KeysView, List, Mapping, Optional,
                    Tuple)
from inspect import getfullargspec
from spinn_utilities.configs import CamelCaseConfigParser
from .log_store import LogStore
from .overrides import overrides

_LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL,
}


class ConfiguredFilter(object):
    """
    Allow a parent logger to filter a child logger.
    """
    __slots__ = [
        "_default_level", "_levels"]

    def __init__(self, config: configparser.RawConfigParser):
        """
        :param config: Parser that read the cfg files
        """
        self._levels = ConfiguredFormatter.construct_logging_parents(config)
        self._default_level = logging.INFO
        if config.has_option("Logging", "default"):
            self._default_level = _LEVELS[config.get("Logging", "default")]

    def filter(self, record: logging.LogRecord) -> bool:
        """
        Get the level for the deepest parent, and filter appropriately.

        :returns: If and only if the log should be done
        """
        level = ConfiguredFormatter.level_of_deepest_parent(
            self._levels, record.name)

        if level is None:
            return record.levelno >= self._default_level

        return record.levelno >= level


class ConfiguredFormatter(logging.Formatter):
    """
    Defines the logging format for the SpiNNaker host software.
    """
    # Precompile this RE; it gets used quite a few times
    __last_component = re.compile(r'\.[^.]+$')

    def __init__(self, config: CamelCaseConfigParser) -> None:
        """
        :param config: Parser that read the cfg files
        """
        if (config.has_option("Logging", "default") and
                config.get("Logging", "default") == "debug"):
            fmt = "%(asctime)-15s %(levelname)s: %(pathname)s: %(message)s"
        else:
            fmt = "%(asctime)-15s %(levelname)s: %(message)s"
        super().__init__(fmt=fmt, datefmt="%Y-%m-%d %H:%M:%S")

    @staticmethod
    def construct_logging_parents(
            config: configparser.RawConfigParser) -> Dict[str, int]:
        """
        Create a dictionary of module names and logging levels.

        This is based on the values if any found in the cfg files.

        :returns: A dictionary of module names and logging levels.
        """
        # Construct the dictionary
        _levels: Dict[str, int] = {}

        if not config.has_section("Logging"):
            return _levels

        for label, level in _LEVELS.items():
            if config.has_option("Logging", label):
                modules = [s.strip() for s in
                           config.get('Logging', label).split(',')]
                if '' not in modules:
                    _levels.update(dict((m, level) for m in modules))
        return _levels

    @staticmethod
    def deepest_parent(parents: KeysView[str], child: str) -> Optional[str]:
        """
        :returns: Greediest match between child and parent.
        """
        # TODO: this can almost certainly be neater!
        # Repeatedly strip elements off the child until we match an item in
        # parents
        match = child

        while '.' in match and match not in parents:
            match = ConfiguredFormatter.__last_component.sub('', match)

        # If no match then return None, there is no deepest parent
        if match not in parents:
            return None

        return match

    @staticmethod
    def level_of_deepest_parent(
            parents: Dict[str, int], child: str) -> Optional[int]:
        """
        :returns:
           The logging level of the greediest match between child and parent.
        """
        # child = re.sub( r'^pacman103\.', '', child )
        parent = ConfiguredFormatter.deepest_parent(parents.keys(), child)

        if parent is None:
            return None

        return parents[parent]


class _BraceMessage(object):
    """
    A message that converts a Python format string to a string.
    """
    __slots__ = [

        "args", "message", "kwargs"]

    def __init__(self, message: object,
                 args: Collection, kwargs: Dict[str, object]) -> None:
        """
        :param message: The log message before formatting
        :param args: Any simple arguments to pass to the formatter
        :param kwargs:Any named arguments to pass to the formatter
        """
        self.message = message
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        try:
            return str(self.message).format(*self.args, **self.kwargs)
        except KeyError:
            try:
                if (not self.args and not self.kwargs and
                        isinstance(self.message, str)):
                    # nothing to format with so assume brackets not formatting
                    return self.message
                return "KeyError: " + str(self.message)
            except KeyError:
                return "Double KeyError"
        except IndexError:
            try:
                if self.args or self.kwargs:
                    return "IndexError: " + str(self.message)
                else:
                    # nothing to format with so assume brackets not formatting
                    return str(self.message)
            except IndexError:
                return "Double IndexError"


class LogLevelTooHighException(Exception):
    """
    An Exception throw when the System tries to log at a level where an
    Exception is a better option.
    """


class FormatAdapter(logging.LoggerAdapter):
    """
    An adaptor for logging with messages that uses Python format strings.

    Example::

        log = FormatAdapter(logging.getLogger(__name__))
        log.info("this message has {} inside {}", 123, 'itself')
        # --> INFO: this message has 123 inside itself
    """
    __kill_level = logging.CRITICAL + 1
    __repeat_at_end = logging.WARNING
    __not_stored_messages: List[Tuple[datetime, int, str]] = []
    __log_store: Optional[LogStore] = None

    @classmethod
    def set_kill_level(cls, level: Optional[int] = None) -> None:
        """
        Allow system to change the level at which a log is changed to an
        Exception.

        .. note::
            This is a static method; it affects all log messages.

        :param level:
            The level to set. The values in :py:mod:`logging` are recommended.
        """
        if level is None:
            cls.__kill_level = logging.CRITICAL + 1
        else:
            cls.__kill_level = level

    @classmethod
    def set_log_store(cls, log_store: Optional[LogStore]) -> None:
        """
        Sets a Object to write the log messages to

        :param log_store:
        """
        if log_store is not None and not isinstance(log_store, LogStore):
            raise TypeError("log_store must be a LogStore")
        cls.__log_store = log_store
        if cls.__log_store:
            for timestamp, level, message in cls._pop_not_stored_messages():
                cls.__log_store.store_log(level, message, timestamp)

    def __init__(
            self, logger: logging.Logger,
            extra: Optional[Mapping[str, object]] = None) -> None:
        """

        :param logger: Logger being wrapped by this adapter
        :param extra:  keyword arguments to pass to the underlying
            standard LoggerAdapter
        """
        if extra is None:
            extra = {}
        super().__init__(logger, extra)
        self.do_log = logger._log  # pylint: disable=protected-access

    @overrides(logging.LoggerAdapter.log, extend_doc=False)
    def log(self, level: int, msg: object,
            *args: object, **kwargs: object) -> None:
        """
        Delegate a log call to the underlying logger, applying appropriate
        transformations to allow the log message to be written using
        Python format string, rather than via `%`-substitutions.
        """
        if level >= FormatAdapter.__kill_level:
            raise LogLevelTooHighException(_BraceMessage(msg, args, kwargs))
        if self.isEnabledFor(level):
            message = _BraceMessage(msg, args, kwargs)
            if FormatAdapter.__log_store:
                try:
                    FormatAdapter.__log_store.store_log(level, str(message))
                except Exception as ex:
                    # Avoid an endless loop of log store errors being logged
                    FormatAdapter.__not_stored_messages.append((
                        datetime.now(),
                        level,
                        f"Unable to store log messages in database due to"
                        f" {ex}"))
                    FormatAdapter.__not_stored_messages.append(
                        (datetime.now(), level, str(message)))
                    FormatAdapter.__log_store = None
                    raise
            else:
                FormatAdapter.__not_stored_messages.append(
                    (datetime.now(), level, str(message)))
            msg, log_kwargs = self.process(msg, kwargs)
            if "exc_info" in kwargs:
                log_kwargs["exc_info"] = kwargs["exc_info"]
            self.do_log(level, message, (), **log_kwargs)

    @overrides(logging.LoggerAdapter.process, extend_doc=False)
    def process(self, msg: object, kwargs: Any) -> Tuple[object, dict]:
        """
        Process the logging message and keyword arguments passed in to a
        logging call to insert contextual information. You can either
        manipulate the message itself, the keyword arguments or both.
        Return the message and *kwargs* modified (or not) to suit your needs.

        :returns: the message and kwargs arguments in both the call
           and the underlying logger.
        """
        return msg, {
            key: kwargs[key]
            for key in getfullargspec(self.do_log).args[1:]
            if key in kwargs}

    @classmethod
    def atexit_handler(cls) -> None:
        """
        Adds code to print out high level log messages python run ends
        """
        messages = []
        if cls.__log_store:
            try:
                messages = cls.__log_store.retreive_log_messages(
                    cls.__repeat_at_end)
            except Exception:  # pylint: disable=broad-except
                # No matter what we don't want an extra Exception reported here
                pass

        messages.extend(map(lambda x: x[2],
                            cls._pop_not_stored_messages(cls.__repeat_at_end)))
        if messages:
            level = logging.getLevelName(cls.__repeat_at_end)
            print(f"\n!WARNING: {len(messages)} log messages were "
                  f"generated at level {level} or above.", file=sys.stderr)
            print("This may mean that the results are invalid.",
                  file=sys.stderr)
            if cls.__log_store:
                print(f"You are advised to check the details of these in "
                      f"the p_log_view of : "
                      f"{cls.__log_store.get_location()}", file=sys.stderr)
            if len(messages) < 10:
                print("These are:", file=sys.stderr)
            else:
                print("The first 10 are:", file=sys.stderr)
            for message in messages[0:10]:
                print(message, file=sys.stderr)

    @classmethod
    def _pop_not_stored_messages(
            cls, min_level: int = 0) -> List[Tuple[datetime, int, str]]:
        """
        Returns the log of messages to print on exit and
        *clears that log*.

        .. note::
            Should only be called externally from test code!
        """
        result: List[Tuple[datetime, int, str]] = []
        for timestamp, level, message in cls.__not_stored_messages:
            if level >= min_level:
                result.append((timestamp, level, message))
        cls.__not_stored_messages = []
        return result


atexit.register(FormatAdapter.atexit_handler)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
from typing import Callable, List, Optional

import appdirs
from typing_extensions import TypeAlias

from spinn_utilities import log
from spinn_utilities.configs import (
    CamelCaseConfigParser, UnexpectedConfigException)

logger = log.FormatAdapter(logging.getLogger(__name__))
_SectionParser: TypeAlias = Callable[[CamelCaseConfigParser], None]


def _check_config(cfg_file: str, default_configs: CamelCaseConfigParser,
                  strict: bool) -> None:
    """
    Checks the configuration read up to this point to see if it is outdated.

    Once one difference is found a full reports is generated and an error
    raised.

    Any section specifically listed as Dead will cause a error

    Any section in the default_cfg should not have extra values.
    It will never have less as the default_cfg are in the configuration.

    Errors on any values listed as PreviousValues.
    These are specific values in specific options no longer supported.
    For example old algorithm names.

    :param cfg_file: Path of last file read in
    :param default_configs:
        configuration with just the default files in
    :param strict: Flag to say an exception should be raised
    """
    if not default_configs.sections():  # empty
        logger.warning("Can not validate cfg files as no default.")
        return
    configs = CamelCaseConfigParser()
    configs.read(cfg_file)
    msg = ""
    for section in configs.sections():
        if default_configs.has_section(section):
            for option in configs.options(section):
                if not default_configs.has_option(section, option):
                    msg += f"Unexpected Option: [{section}]{option}\n"
        else:
            msg += f"Unexpected Section: [{section}]\n"
    if msg:
        msg += f"found in {cfg_file}"
        if strict:
            raise UnexpectedConfigException(msg)
        else:
            logger.warning(msg)


def _read_a_config(
        configuration: CamelCaseConfigParser, cfg_file: str,
        default_configs: CamelCaseConfigParser, strict: bool) -> None:
    """
    Reads in a configuration file and then directly its `machine_spec_file`.

    :param configuration:
        configuration to be updated by the reading of a file
    :param cfg_file: path to file which should be read in
    :param default_configs:
        configuration with just the default files in
    :param strict: Flag to say checker should raise an exception
    """
    _check_config(cfg_file, default_configs, strict)
    configuration.read(cfg_file)
    if configuration.has_option("Machine", "machine_spec_file"):
        machine_spec_file = configuration.get("Machine", "machine_spec_file")
        _check_config(machine_spec_file, default_configs, strict)
        configuration.read(machine_spec_file)
        configuration.remove_option("Machine", "machine_spec_file")


def _config_locations(filename: str) -> List[str]:
    """
    Defines the list of places we can get configuration files from.

    :param filename:
        The local name of the configuration file, e.g., 'spynnaker.cfg'
    :return: list of fully-qualified filenames
    """
    dotname = "." + filename

    # locations to read as well as default later overrides earlier
    system_config_cfg_file = os.path.join(appdirs.site_config_dir(), dotname)
    user_config_cfg_file = os.path.join(appdirs.user_config_dir(), dotname)
    user_home_cfg_file = os.path.join(os.path.expanduser("~"), dotname)

    # locations to read as well as default later overrides earlier
    return [system_config_cfg_file, user_config_cfg_file,
            user_home_cfg_file]


def load_defaults(defaults: List[str]) -> CamelCaseConfigParser:
    """
    Load the default configuration.

    :param defaults:
        The list of files to get default configurations from.
    :return: the fully-loaded configuration
    """
    default_configs = CamelCaseConfigParser()
    default_configs.read(defaults)
    return default_configs


def load_config(local_name: Optional[str], user_cfg: Optional[str],
                defaults: List[str]) -> CamelCaseConfigParser:
    """
    Load the configuration.

    :param local_name:
        Name of the file to look for in the current directory
    :param user_cfg:
        Path to existing user cfg. This file must exist.
    :param defaults:
        The list of files to get default configurations from.
    :return: the fully-loaded and checked configuration
    """
    configs = load_defaults(defaults)
    default_configs = load_defaults(defaults)

    if user_cfg is not None:
        _read_a_config(configs, user_cfg, default_configs, False)
    if local_name is not None:
        local_path = os.path.join(os.curdir, local_name)
        _read_a_config(configs, local_path, default_configs, True)

    # Log which configs files we read
    logger.info("Read configs files: {}", ", ".join(configs.read_files))

    return configs



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import FunctionType, MethodType
from typing import Any, Callable, TypeVar

#: :meta private:
Method = TypeVar("Method", bound=Callable[..., Any])


#  pylint: disable=invalid-name
class overrides(object):
    """
    A decorator for indicating that a method overrides another method in
    a superclass.  This checks that the method does actually exist,
    copies the doc-string for the method, and enforces that the method
    overridden is specified, making maintenance easier.
    """
    __slots__ = [
        # The method in the superclass that this method overrides
        "_superclass_method",
        # True if the doc string is to be extended, False to set if not set
        "_extend_doc",
    ]

    def __init__(
            self, super_class_method: Callable, *, extend_doc: bool = True):
        """
        :param super_class_method: The method to override in the superclass
        :param extend_doc:
            True the method doc string should be appended to the super-method
            doc string, False if the documentation should be set to the
            super-method doc string only if there isn't a doc string already
        """
        if isinstance(super_class_method, property):
            super_class_method = super_class_method.fget
        if not isinstance(super_class_method, (FunctionType, MethodType)):
            raise TypeError("may only decorate method declarations; "
                            f"this is a {type(super_class_method)}")
        self._superclass_method = super_class_method
        self._extend_doc = bool(extend_doc)

    def __call__(self, method: Method) -> Method:
        """
        Apply the decorator to the given method.
        """
        # Check and fail if this is a property
        if isinstance(method, property):
            raise AttributeError(
                f"Please ensure that the {self.__class__.__name__} decorator "
                "is the last decorator before the method declaration")

        # Check that the name matches
        if (method.__name__ != self._superclass_method.__name__):
            raise AttributeError(
                f"Name mismatch: {self._superclass_method.__name__} "
                f" !=- {method.__name__}")

        if (self._superclass_method.__doc__ is not None and
                method.__doc__ is None):
            method.__doc__ = self._superclass_method.__doc__
        elif (self._extend_doc and
                self._superclass_method.__doc__ is not None):
            method.__doc__ = (
                self._superclass_method.__doc__ + (method.__doc__ or ""))
        return method



# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import yaml
import io
import importlib
import argparse
from types import ModuleType
from typing import Any, Dict, List, Optional, Set, Union
import sys
from .citation_updater_and_doi_generator import CitationUpdaterAndDoiGenerator

ENCODING = "utf-8"

REQUIREMENTS_FILE = "requirements.txt"
C_REQUIREMENTS_FILE = "c_requirements.txt"
CITATION_FILE = "CITATION.cff"
PYPI_TO_IMPORT_FILE = "pypi_to_import"

REFERENCES_YAML_POINTER = "references"
REFERENCE_TYPE = "software"
REFERENCES_AUTHORS_TYPE = "authors"
REFERENCES_TITLE_TYPE = "title"
REFERENCES_VERSION_TYPE = "version"
REFERENCES_DATE_TYPE = "date-released"
REFERENCES_URL_TYPE = "url"
REFERENCES_REPO_TYPE = "repository"
REFERENCES_CONTACT_TYPE = "contact"
REFERENCES_TYPE_TYPE = "type"
REFERENCES_SOFTWARE_TYPE = "software"

CITATION_DOI_TYPE = 'identifier'

# pylint: skip-file

_SEEN_TYPE = Set[Union[ModuleType, str, None]]


class CitationAggregator(object):
    """
    Helper class for building a citation file which references all
    dependencies.
    """

    def create_aggregated_citation_file(
            self, module_to_start_at: ModuleType,
            aggregated_citation_file: str) -> None:
        """
        Entrance method for building the aggregated citation file.

        :param module_to_start_at:
            the top level module to figure out its citation file for
        :param aggregated_citation_file:
            file name of aggregated citation file
        """
        # get the top citation file to add references to
        module_file: Optional[str] = module_to_start_at.__file__
        assert module_file is not None
        top_citation_file_path = os.path.join(os.path.dirname(os.path.dirname(
            os.path.abspath(module_file))), CITATION_FILE)
        modules_seen_so_far: _SEEN_TYPE = set()
        modules_seen_so_far.add("")
        with open(top_citation_file_path, encoding=ENCODING) as stream:
            top_citation_file: Dict[str, Any] = yaml.safe_load(
                stream)
        top_citation_file[REFERENCES_YAML_POINTER] = list()

        # get the dependency list
        requirements_file_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(
                module_file))), REQUIREMENTS_FILE)
        c_requirements_file_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(
                module_file))), C_REQUIREMENTS_FILE)

        # attempt to get python PYPI to import command map
        pypi_to_import_map_file = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(
                module_file))), PYPI_TO_IMPORT_FILE)
        pypi_to_import_map = None
        if os.path.isfile(pypi_to_import_map_file):
            pypi_to_import_map = self._read_pypi_import_map(
                pypi_to_import_map_file)

        if os.path.isfile(requirements_file_path):
            with open(requirements_file_path, encoding="utf-8") as r_file:
                for line in r_file:
                    module = line.strip()
                    if module.startswith("#"):
                        continue
                    if module not in modules_seen_so_far:
                        assert pypi_to_import_map is not None
                        import_name = pypi_to_import_map.get(module, module)
                        #  pylint: disable=broad-except
                        try:
                            imported_module = importlib.import_module(
                                import_name)
                            self._handle_python_dependency(
                                top_citation_file, imported_module,
                                modules_seen_so_far,
                                pypi_to_import_map[module])
                        except Exception as e:  # pragma: no cover
                            print("Error handling python dependency "
                                  f"{module}: {e}")

        if os.path.isfile(c_requirements_file_path):
            with open(c_requirements_file_path, encoding=ENCODING) as r_file:
                for line in r_file:
                    module = line.strip()
                    if module.startswith("#"):
                        continue
                    if module not in modules_seen_so_far:
                        self._handle_c_dependency(
                            top_citation_file, module, modules_seen_so_far)

        # write citation file with updated fields
        with io.open(
                aggregated_citation_file, 'w', encoding=ENCODING) as outfile:
            yaml.dump(top_citation_file, outfile, default_flow_style=False,
                      allow_unicode=True)

    @staticmethod
    def _read_pypi_import_map(aggregated_citation_file: str) -> Dict[str, str]:
        """
        Read the PYPI to import name map.

        :param aggregated_citation_file: path to the PYPI map file
        :return: map between PYPI names and import names
        """
        pypi_to_import_map: Dict[str, str] = dict()
        with open(aggregated_citation_file, encoding=ENCODING) as f:
            for line in f:
                [pypi, import_command] = line.split(":")
                pypi_to_import_map[pypi] = import_command.split("\n")[0]
        return pypi_to_import_map

    def _handle_c_dependency(
            self, top_citation_file:  Dict[str, Any], module: str,
            modules_seen_so_far: _SEEN_TYPE) -> None:
        """
        Handle a C code dependency.

        :param top_citation_file: YAML file for the top citation file
        :param module: module to find
        :param modules_seen_so_far:
        """
        cleaned_path = self.locate_path_for_c_dependency(module)
        if cleaned_path is not None:
            # process reference
            reference_entry = self._process_reference(
                cleaned_path, None, modules_seen_so_far, module)

            # append to the top citation file
            top_citation_file[REFERENCES_YAML_POINTER].append(
                reference_entry)
            self._search_for_other_c_references(
                reference_entry, cleaned_path, modules_seen_so_far)
        else:
            print(f"Could not find C dependency {module}")

    @staticmethod
    def locate_path_for_c_dependency(true_software_name: str) -> Optional[str]:
        """
        Tries to find the software in the environment PATH (s)

        :returns: Path to the software if found
        """
        environment_path_variable = os.environ.get('PATH')
        if environment_path_variable is not None:
            software_paths = environment_path_variable.split(":")
            for software_path in software_paths:
                # clear path to have repository name at end
                last_version = None
                cleaned_path = software_path
                while ((cleaned_path != last_version) and (
                        not (cleaned_path.split(os.sep)[-1] ==
                             true_software_name))):
                    last_version = cleaned_path
                    cleaned_path = os.path.dirname(cleaned_path)

                if cleaned_path != last_version:
                    return cleaned_path
        return None

    def _search_for_other_c_references(
            self, reference_entry:  Dict[str, Any], software_path: str,
            modules_seen_so_far: _SEEN_TYPE) -> None:
        """
        Go through the top level path and tries to locate other CFF
        files that need to be added to the references pile.

        :param reference_entry:
            The reference entry to add new dependencies as references for.
        :param software_path: the path to search in
        :param modules_seen_so_far:
        """
        for possible_extra_citation_file in os.listdir(software_path):
            if possible_extra_citation_file.endswith(".cff"):
                dependency_reference_entry = \
                    self._read_and_process_reference_entry(
                        os.path.join(software_path,
                                     possible_extra_citation_file))
                reference_entry[REFERENCES_YAML_POINTER] = list()
                reference_entry[REFERENCES_YAML_POINTER].append(
                    dependency_reference_entry)
                modules_seen_so_far.add(
                    possible_extra_citation_file.split(".")[0])

    def _handle_python_dependency(
            self, top_citation_file: Dict[str, Any],
            imported_module: ModuleType, modules_seen_so_far: _SEEN_TYPE,
            module_name: str) -> None:
        """
        Handle a python dependency.

        :param top_citation_file:
            YAML file for the top citation file
        :param imported_module: the actual imported module
        :param modules_seen_so_far:
            list of names of dependencies already processed
        :param module_name:
            the name of this module to consider as a dependency
        :raises FileNotFoundError:
        """
        # get modules citation file
        module_path = imported_module.__file__
        assert module_path is not None
        citation_level_dir = os.path.abspath(module_path)
        m_path = module_name.replace(".", os.sep)
        last_citation_level_dir = None
        while (not citation_level_dir.endswith(m_path) and
               last_citation_level_dir != citation_level_dir):
            last_citation_level_dir = citation_level_dir
            citation_level_dir = os.path.dirname(citation_level_dir)
        if citation_level_dir == last_citation_level_dir:  # pragma: no cover
            raise FileNotFoundError(
                f"Folder for module {module_name} not found")

        # get the reference data for the reference
        reference_entry = self._process_reference(
            citation_level_dir, imported_module, modules_seen_so_far,
            module_name)

        if reference_entry is not None:
            # append to the top citation file
            top_citation_file[REFERENCES_YAML_POINTER].append(reference_entry)

    def _process_reference(
            self, citation_level_dir: str,
            imported_module: Optional[ModuleType],
            modules_seen_so_far: _SEEN_TYPE,
            module_name: str) -> Dict[str, Any]:
        """
        Take a module level and tries to locate and process a citation file.

        :param citation_level_dir:
            the expected level where the ``CITATION.cff`` should be
        :param imported_module: the module after being imported
        :param modules_seen_so_far:
            list of dependencies already processed
        :return: the reference entry in JSON format
        """
        # if it exists, add it as a reference to the top one
        if os.path.isfile(os.path.join(citation_level_dir, CITATION_FILE)):
            reference_entry = self._read_and_process_reference_entry(
                os.path.join(citation_level_dir, CITATION_FILE))

        # check that the file isn't one above (not installed, but developer
        # mode)
        elif os.path.isfile(os.path.join(os.path.dirname(
                os.path.abspath(citation_level_dir)), CITATION_FILE)):
            reference_entry = self._read_and_process_reference_entry(
                os.path.join(os.path.dirname(
                    os.path.abspath(citation_level_dir)), CITATION_FILE))

        # if no citation file exists, do an attempt to find a version to build
        # from
        else:
            # one from version
            reference_entry = self._try_to_find_version(
                imported_module, module_name)

        modules_seen_so_far.add(imported_module)
        return reference_entry

    @staticmethod
    def _try_to_find_version(
            imported_module: Optional[ModuleType],
            module_name: str) -> Dict[str, Any]:
        """
        Try to locate a version file or version data to auto-generate
        minimal citation data.

        :param imported_module:
            the module currently trying to find the version of
        :return: reference entry for this python module
        """
        reference_entry: Dict[str, Any] = dict()
        reference_entry[REFERENCES_TYPE_TYPE] = REFERENCES_SOFTWARE_TYPE
        reference_entry[REFERENCES_TITLE_TYPE] = module_name
        if imported_module is None:
            return reference_entry
        if (hasattr(imported_module, "__version_day__") and
                hasattr(imported_module, "__version_month__") and
                hasattr(imported_module, "__version_year__")):
            reference_entry[REFERENCES_DATE_TYPE] = \
                CitationUpdaterAndDoiGenerator.\
                convert_text_date_to_date(
                    version_day=imported_module.__version_day__,
                    version_month=imported_module.__version_month__,
                    version_year=imported_module.__version_year__)
        if hasattr(imported_module, "__version__"):
            reference_entry[REFERENCES_VERSION_TYPE] = \
                imported_module.__version__
        elif hasattr(imported_module, "version"):
            reference_entry[REFERENCES_VERSION_TYPE] = \
                imported_module.version
        elif hasattr(imported_module, "_version"):
            # pylint: disable=protected-access
            reference_entry[REFERENCES_VERSION_TYPE] = \
                imported_module._version
        return reference_entry

    @staticmethod
    def _read_and_process_reference_entry(
            dependency_citation_file_path: str) -> Dict[str, Any]:
        """
        Read a ``CITATION.cff`` and makes it a reference for a higher
        level citation file.

        :param dependency_citation_file_path:
            path to a `CITATION.cff` file
        :return: reference entry for the higher level `CITATION.cff`
        """
        reference_entry = dict()

        with open(dependency_citation_file_path, 'r', encoding="utf-8") \
                as stream:
            dependency_citation_file = yaml.safe_load(stream)

            reference_entry[REFERENCES_TYPE_TYPE] = REFERENCES_SOFTWARE_TYPE
            reference_entry[REFERENCES_AUTHORS_TYPE] = \
                dependency_citation_file[REFERENCES_AUTHORS_TYPE]
            reference_entry[REFERENCES_TITLE_TYPE] = \
                dependency_citation_file[REFERENCES_TITLE_TYPE]
            reference_entry[REFERENCES_CONTACT_TYPE] = \
                dependency_citation_file[REFERENCES_CONTACT_TYPE]
            reference_entry[REFERENCES_VERSION_TYPE] = \
                dependency_citation_file[REFERENCES_VERSION_TYPE]
            reference_entry[REFERENCES_DATE_TYPE] = \
                dependency_citation_file[REFERENCES_DATE_TYPE]
            reference_entry[REFERENCES_URL_TYPE] = \
                dependency_citation_file[REFERENCES_URL_TYPE]
            reference_entry[REFERENCES_REPO_TYPE] = \
                dependency_citation_file[REFERENCES_REPO_TYPE]
        return reference_entry


def generate_aggregate(arguments: Optional[List[str]] = None) -> None:
    """
    Command-line tool to generate a single ``citation.cff`` from others.

    :param arguments: Command line arguments.

        * ``--output_path``: \
            Where to write the aggregate file
        * ``--top_module``: \
            The module to start aggregating the ``citation.cff``\\s from
        * ``--doi_title``: \
            The title of the DOI
        * ``--zenodo_access_token``: \
            The access token for Zenodo
        * ``--tools_doi``: \
            The DOI of the tools
    """
    parser = argparse.ArgumentParser(description="Aggregate Citations")
    parser.add_argument("output_path", help="The file to store the result in")
    parser.add_argument("top_module", help="The module to start with")
    parser.add_argument("--create_doi", action="store_true",
                        help="Create a DOI from the resulting citation"
                             " on Zenodo")
    parser.add_argument("--publish_doi", action="store_true",
                        help="Publish the DOI created")
    parser.add_argument("--doi_title",
                        help="The title to give the created DOI")
    parser.add_argument("--previous_doi",
                        help="The DOI this is a newer version of")
    parser.add_argument("--zenodo_access_token",
                        help="Access token for Zenodo")

    args = parser.parse_args(arguments)
    error = False
    if args.create_doi:  # pragma: no cover
        if not args.doi_title:
            print("--doi_title required when creating a DOI")
            error = True
        if not args.previous_doi:
            print("--previous_doi required when creating a DOI")
            error = True
        if not args.zenodo_access_token:
            print("--zenodo_access_token required when creating a DOI")
            error = True
    if args.publish_doi and not args.create_doi:  # pragma: no cover
        print("Cannot publish DOI without creating one")
        error = True
    if error:  # pragma: no cover
        parser.print_usage()
        sys.exit()

    top_module = importlib.import_module(args.top_module)

    citation_aggregator = CitationAggregator()
    citation_aggregator.create_aggregated_citation_file(
        top_module, args.output_path)
    citation_updater_and_dio_generator = CitationUpdaterAndDoiGenerator()
    citation_updater_and_dio_generator.update_citation_file_and_create_doi(
        citation_file_path=args.output_path,
        doi_title=args.doi_title,
        create_doi=args.create_doi, publish_doi=args.publish_doi,
        previous_doi=args.previous_doi,
        zenodo_access_token=args.zenodo_access_token,
        module_path=os.path.dirname(top_module.__path__[0]))


if __name__ == "__main__":
    generate_aggregate()  # pragma: no cover



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sized
import logging
import math
import os
import sys
from types import TracebackType
from typing import Iterable, Optional, Type, TypeVar, Union

from typing_extensions import Literal, Self

from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_utilities import logger_utils


logger = FormatAdapter(logging.getLogger(__name__))
#: :meta private:
T = TypeVar("T")


class ProgressBar(object):
    """
    Progress bar for telling the user where a task is up to.
    """
    MAX_LENGTH_IN_CHARS = 60

    TOO_MANY_ERROR = (
        "Too many update steps in progress bar! "
        "This may be a sign that something else has gone wrong!")

    __slots__ = (
        "_number_of_things", "_currently_completed", "_destination",
        "_chars_per_thing", "_chars_done", "_string",
        "_step_character", "_end_character", "_in_bad_terminal",
    )

    def __init__(self, total_number_of_things_to_do: Union[int, Sized],
                 string_describing_what_being_progressed: str,
                 step_character: str = "=", end_character: str = "|"):
        """
        :param total_number_of_things_to_do: number of steps being shown..
           Will use the length of any object if possible.
        :param string_describing_what_being_progressed: text for the bar
        :param step_character: Optional: characters to print
        :param end_character: Optional: characters to print as start and end
        """
        if isinstance(total_number_of_things_to_do, Sized):
            self._number_of_things = len(total_number_of_things_to_do)
        else:
            self._number_of_things = int(total_number_of_things_to_do)
        self._currently_completed = 0
        self._chars_per_thing = 1.0
        self._chars_done = 0
        self._string = string_describing_what_being_progressed
        self._destination = sys.stderr
        self._step_character = step_character
        self._end_character = end_character

        # Determine if we are in a "bad" terminal i.e. one that doesn't handle
        # carriage return correctly
        self._in_bad_terminal = "PROGRESS_GOOD_TERMINAL" not in os.environ

        self._create_initial_progress_bar(
            string_describing_what_being_progressed)

    def update(self, amount_to_add: int = 1) -> None:
        """
        Update the progress bar by a given amount.

        :param amount_to_add:
        """
        if self._currently_completed + amount_to_add > self._number_of_things:
            logger_utils.error_once(logger, self.TOO_MANY_ERROR)
            return
        self._currently_completed += amount_to_add
        self._check_differences()

    def _print_overwritten_line(self, string: str) -> None:
        print("\r" + string, end="", file=self._destination)

    def _print_distance_indicator(self, description: str) -> None:
        if description is not None:
            print(description, file=self._destination)

        # Find the mid point
        mid_point = ProgressBar.MAX_LENGTH_IN_CHARS // 2

        # The space between 0% and 50% is the mid-point minus the width of
        # 0% and ~half the width of 50%
        first_space = mid_point - 4

        # The space between 50% and 100% is the mid-point minus the rest of
        # the width of 50% and the width of 100%
        second_space = mid_point - 5

        # Print the progress bar itself
        self._print_distance_line(first_space, second_space)
        if self._in_bad_terminal:
            print("", file=self._destination)
            print(" ", end="", file=self._destination)

    def _print_distance_line(
            self, first_space: int, second_space: int) -> None:
        line = f"{self._end_character}0%{' ' * first_space}50%" \
               f"{' ' * second_space}100%{self._end_character}"
        print(line, end="", file=self._destination)

    def _print_progress(self, length: int) -> None:
        chars_to_print = length
        if not self._in_bad_terminal:
            self._print_overwritten_line(self._end_character)
        else:
            chars_to_print = length - self._chars_done

        for _ in range(int(chars_to_print)):
            self._print_progress_unit()
        self._destination.flush()

    def _print_progress_unit(self) -> None:
        print(self._step_character, end='', file=self._destination)

    def _print_progress_done(self) -> None:
        self._print_progress(ProgressBar.MAX_LENGTH_IN_CHARS)
        if not self._in_bad_terminal:
            print(self._end_character, file=self._destination)
        else:
            print("", file=self._destination)

    def _create_initial_progress_bar(self, description: str) -> None:
        if self._number_of_things == 0:
            self._chars_per_thing = ProgressBar.MAX_LENGTH_IN_CHARS
        else:
            self._chars_per_thing = \
                ProgressBar.MAX_LENGTH_IN_CHARS / self._number_of_things
        self._print_distance_indicator(description)
        self._print_progress(0)
        self._check_differences()

    def _check_differences(self) -> None:
        expected_chars_done = int(math.floor(
            self._currently_completed * self._chars_per_thing))
        if self._currently_completed == self._number_of_things:
            expected_chars_done = ProgressBar.MAX_LENGTH_IN_CHARS
        self._print_progress(expected_chars_done)
        self._chars_done = expected_chars_done

    def end(self) -> None:
        """
        Close the progress bar, updating whatever is left if needed.
        """
        difference = self._number_of_things - self._currently_completed
        self._currently_completed += difference
        self._check_differences()
        self._print_progress_done()

    def __repr__(self) -> str:
        return f"<ProgressBar:{self._string}>"

    def __enter__(self) -> Self:
        """
        Support method to use the progress bar as a context manager::

            with ProgressBar(...) as p:
                ...
                p.update()
                ...
                p.update()
                ...

        This method does not have any parameters because any parameters in the
        with :samp:`ProgressBar(...)` call have been passed to
        :py:meth:`__init__`

        Like :samp:`__new__` this method has to return self as in theory it
        could pass back a different object. Welcome to Python.

        :return: The Progress bar
        """
        return self

    def __exit__(self, exc_type: Optional[Type], exc_val: Exception,
                 exc_tb: TracebackType) -> Literal[False]:
        self.end()
        return False

    def over(self, collection: Iterable[T],
             finish_at_end: bool = True) -> Iterable[T]:
        """
        Simple wrapper for the cases where the progress bar is being used
        to show progress through the iteration over a single collection.
        The progress bar should have been initialised to the size of the
        collection being iterated over.

        :param collection:
            The base collection (any iterable) being iterated over
        :param finish_at_end:
            Flag to say if the bar should finish at the end of the collection
        :return: An iterable. Expected to be directly used in a for.
        """
        try:
            for item in collection:
                yield item
                self.update()
        finally:
            if finish_at_end:
                self.end()


class DummyProgressBar(ProgressBar):
    """
    This is a dummy version of the progress bar that just stubs out the
    internal printing operations with code that does nothing. It otherwise
    fails in exactly the same way.
    """
    @overrides(ProgressBar._print_overwritten_line)
    def _print_overwritten_line(self, string: str) -> None:
        pass

    @overrides(ProgressBar._print_distance_indicator)
    def _print_distance_indicator(self, description: str) -> None:
        pass

    @overrides(ProgressBar._print_progress)
    def _print_progress(self, length: int) -> None:
        pass

    @overrides(ProgressBar._print_progress_done)
    def _print_progress_done(self) -> None:
        pass

    def __repr__(self) -> str:
        return f"<DummyProgressBar:{self._string}>"


if __name__ == "__main__":  # pragma: no cover
    from time import sleep
    demo = ProgressBar(
        5, "Progress Bar Demonstration", step_character="-", end_character="!")
    for _ in range(5):
        sleep(1)
        demo.update()
    demo.end()
    demo = ProgressBar(30, "Progress Bar Demonstration")
    for _ in range(30):
        sleep(0.1)
        demo.update()
    demo.end()

    _collection = [2, 3, 5, 7, 11, 13, 17]
    demo = ProgressBar(_collection, "Demo over a few primes")
    for prime in demo.over(_collection):
        sleep(0.1)

    demo = DummyProgressBar(3, "Test Dummy")
    for prime in demo.over(_collection):
        sleep(0.1)



# Copyright (c) 2025 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import os
import re
import sys
from typing import Dict, List, Set, TextIO

from spinn_utilities.config_holder import (
    get_default_cfgs)
from spinn_utilities.configs.camel_case_config_parser import (
    TypedConfigParser)


def _make_name(option: str) -> str:
    """
    Converts an option to a possibly shortened name string

    This will no include a leading #

    This is used for cfg options that may be automatically group.
    So that either can be used in a link.

    :return: Possibly shortened option name
    """
    # take the link marker out
    if option.startswith("#"):
        option = option[1:]
    if option.startswith("@"):
        raise ValueError(f"{option=} has a @")
    elif option.startswith("draw_"):
        option = option[5:]
    elif option.startswith("keep_"):
        option = option[5:]
    elif option.startswith("path_"):
        option = option[5:]
    elif option.startswith("tpath_"):
        option = option[6:]
    elif option.startswith("run_"):
        option = option[4:]
    elif option.startswith("write_"):
        option = option[6:]
    return option


def _make_link(option: str) -> str:
    """
    Converts an option to a possibly shortened link string

    This will include a leading #

    see _make_name

    :return: Possibly shortened option link
    """
    return "#" + _make_name(option)


def _md_write_doc(f: TextIO, raw: str) -> None:
    """
    Writes the raw (document text) with some minor corrections.

    See
    https://github.com/SpiNNakerManchester/SpiNNUtils/spinn_utilities/configs/notes.md
    for the changes that will be made.
    """
    while "\\t*" in raw:
        raw = raw.replace("\\t*", "   *")
    while "\\t" in raw:
        raw = raw.replace("\\t", "&nbsp;&nbsp;")
    if raw[-2:] == "\\n":
        raw = raw[:-2] + "\n"
    raw = raw.replace("\\n\\n", "\n\n")
    raw = raw.replace("\\n\n", "  \n")

    pointers = []
    pointer_matchs = re.finditer(r"\[(\w|\s|\(|\))*\]\((\s|\w)*\)", raw)
    for pointer_match in pointer_matchs:
        pointers.append(raw[pointer_match.start():pointer_match.end()])
    for pointer in pointers:
        start = pointer.index("]", ) + 2
        raw_link = pointer[start:-1]
        link = _make_link(raw_link)
        if link != raw_link:
            fixed = pointer[:start] + link + ")"
            raw = raw.replace(pointer, fixed)

    f.write(raw)
    f.write("\n\n")


class _ConfigGroup(object):
    """
    Hold the values for one cfg key plus its docs and any grouped ones.
    """

    def __init__(self, option: str, value: str):
        """
        :param option: cfg option name
        :param value: value assigned to this option in the cfg file
        """
        self._docs: str = ""  # Must be changed later
        self.title = option
        self._cfg: Dict[str, str] = dict()
        self._cfg[option] = value

    def paths(self) -> List[str]:
        """
        Gets a list of the cfg settings in the group that point to a path.

        Path cfg options are those which start with path_ or tpath_

        :returns: The List of cfg values which point to a path
        """
        paths = []
        for option, value in self._cfg.items():
            if option.startswith("path_"):
                paths.append(value)
            if option.startswith("tpath_"):
                paths.append(value)
        return paths

    def print_config(self) -> None:
        """
        Prints the keys and Values in this cfg group.

        This is mainly for debugging so can change without notice.
        """
        if len(self._cfg) == 1:
            for option, value in self._cfg.items():
                print(f"\t{option}: {value}")
        else:
            print(f"\t{self.title}")
            for option, value in self._cfg.items():
                print(f"\t\t{option}: {value}")

    def merge(self, other: "_ConfigGroup") -> None:
        """
        Merges the option form the other into this group.
        """
        # pylint: disable=protected-access
        for option, value in other._cfg.items():
            if option in self._cfg:
                raise ValueError(f"{option} already exists in group")
            self._cfg[option] = value

    def add_doc(self, docs: str) -> None:
        """
        Adds the docs (marked with @) into this group
        """
        if self._docs:
            raise ValueError(f"{self.title} already has a docs")
        self._docs = docs

    def missing_docs(self) -> bool:
        """
        :returns: True if this group does not have any docs
        """
        return not self._docs

    def _md_value(self, value: str) -> str:
        """
        Writes the value adding a pointer to mode if applicable.
        """
        if value.lower() in ["debug", "info"]:
            if self.title == "default":
                # special case [Logging]default
                return value
            else:
                return f"[{value}](#mode)"
        else:
            return value

    def md(self, f: TextIO) -> None:
        """
        Writes a markdown version of this group to the file.
        """
        f.write(f'### <a name="{_make_name(self.title)}"></a> {self.title}\n')
        if len(self._cfg) == 1:
            value = next(iter(self._cfg.values()))
            f.write(f"Default: __{self._md_value(value)}__\n")
        else:
            for option, value in self._cfg.items():
                f.write(f"* key: {option} \n")
                f.write(f"  * value: {self._md_value(value)}\n")
        f.write("\n")
        _md_write_doc(f, self._docs)


class ConfigDocumentor(object):
    """
    This class will document all the default configs.

    This can be called at any level but will only work with the
    default cfg files set. (Typically up to that level)
    """

    def __init__(self) -> None:
        self._sections: Dict[str, Dict[str, _ConfigGroup]] = defaultdict(dict)
        self._docs: Dict[str, str] = dict()
        self._names: Dict[str, str] = dict()
        config1 = TypedConfigParser()
        config1.read(get_default_cfgs())
        for section in config1:
            if section == "DEFAULT":
                continue
            for option in config1.options(section):
                value = config1.get(section, option)
                self._add_option(section, option, value)
        self._process_special(config1)
        self._merge_paths()
        self.check()

    def _add_option(self, section: str, option: str, value: str) -> None:
        """
        Adds a cfg option.

        Ignores the special ones (@) at this time.
        """
        groups = self._sections[section]
        if option.startswith("@"):
            return
        if option in groups:
            raise ValueError(f"{option=} used twice")
        else:
            group = _ConfigGroup(option, value)
            groups[option] = group
        name = _make_name(option)
        if name != option:
            if option.startswith("path_") or option.startswith("tpath_"):
                return
            if name in self._names:
                raise ValueError(f"Both {option} and {self._names[name]}"
                                 f" have the same link.")
            self._names[name] = option

    def _process_special(self, config: TypedConfigParser) -> None:
        """
        This pass processes all the special (start with @) cfg keys.

        See
        https://github.com/SpiNNakerManchester/SpiNNUtils/spinn_utilities/configs/notes.md
        """
        for section in config.sections():
            groups = self._sections[section]
            for option in config.options(section):
                if option.startswith("@"):
                    if option.startswith("@group"):
                        if option.startswith("@group_"):
                            group_option = option[7:]
                        else:
                            group_option = option[6:]
                        other_option = config.get(section, option)
                        groups[other_option].merge(groups[group_option])
                        del self._sections[section][group_option]
                    else:
                        if option == "@":
                            self._docs[section] = config.get(section, option)
                        else:
                            other_option = option[1:]
                            groups[other_option].add_doc(config.get(
                                section, option))

    def _merge_paths(self) -> None:
        """
        Merges groups with path and similar option names.

        See
        https://github.com/SpiNNakerManchester/SpiNNUtils/spinn_utilities/configs/notes.md
        """
        for groups in self._sections.values():
            remove_options = set()
            for option in groups:
                if option.startswith("path_") or option.startswith("tpath_"):
                    as_link = _make_name(option)
                    if as_link in self._names:
                        other_option = self._names[as_link]
                        groups[other_option].merge(groups[option])
                        remove_options.add(option)
            for option in remove_options:
                del groups[option]

    def check(self) -> None:
        """
        Runs the checks that the cfg will work as markdown.

        Expects all sections and not grouped (including by path name) options
        to have docs (@)

        Expects all  sections and not grouped options to generate unique links.
        """
        names: Set[str] = set()
        for section in self._sections:
            if section in names:
                raise ValueError(f"{section=} has a duplicate name")
            else:
                names.add(section)
            if section not in self._docs:
                raise ValueError(f"{section=} has no doc")
            for title, group in self._sections[section].items():
                name = _make_name(title)
                if name in names:
                    raise ValueError(
                        f"{name=} from {title=} has a duplicate link")
                else:
                    names.add(name)
                if group.missing_docs():
                    raise ValueError(f"{title=} has missing docs")

    def print_section(self, section: str) -> None:
        """
        Prints this section.

        This is mainly for debugging so can change without notice.
        """
        print(section)
        print(f"\t{self._docs[section]}")
        titles = list(self._sections[section])
        titles.sort()
        groups = self._sections[section]
        for title in titles:
            group = groups[title]
            group.print_config()

    def print_configs(self) -> None:
        """
        Prints all configs.

        This is mainly for debugging so can change without notice.
        """

        for section in self._sections:
            self.print_section(section)

    def _md_header(self, f: TextIO) -> None:
        """
        Write the mark down header section.
        """
        f.write("---\n")
        f.write("option: SpiNNaker cfg settings\n")
        f.write("layout: default\n")
        f.write("published: true\n")
        f.write("---\n")
        f.write("CFG Settings and Reports\n")
        f.write("========================\n")
        f.write("<!-- \n")
        f.write("This file is autogenerated do not edit!\n")
        f.write("Created by spinn_utilities/configs/config_documentor.py\n")
        f.write("Based on the default cfg files\n")
        f.write("See notes section at the bottom of this file.\n")
        f.write("-->\n")
        f.write("This guide covers the cfg settings "
                "and the reports created.\n")

    def _md_section(self, section: str, f: TextIO) -> None:
        """
        Writes the markdown for one section.
        """
        f.write(f'# <a name="{section}"></a> {section}\n')
        if section in self._docs:
            _md_write_doc(f, self._docs[section])
        titles = list(self._sections[section])
        titles.sort()
        for title in titles:
            f.write(f"* [{title}](#{title})\n")
        f.write("\n")
        for title in titles:
            group = self._sections[section][title]
            group.md(f)

    def md_reports(self, f: TextIO) -> None:
        """
        Writes in markdown format a list of links from paths to cfg options.
        """
        f.write('# <a name="report_files"></a> Report Files\n')
        p_map = dict()
        for section in self._sections:
            for group in self._sections[section].values():
                for path in group.paths():
                    p_map[path] = _make_link(group.title)
        p_map = dict(sorted(p_map.items()))
        for path, title in p_map.items():
            f.write(f"  * [{path}](#{title})\n")
        f.write("\n\n")

    def md_notes(self, f: TextIO) -> None:
        """
        Copied the notes file into this markdown file
        """
        class_file = sys.modules[self.__module__].__file__
        assert class_file is not None
        abs_class_file = os.path.abspath(class_file)
        class_dir = os.path.dirname(abs_class_file)
        notes_path = os.path.join(class_dir, 'notes.md')
        with open(notes_path, encoding="utf-8") as notesfile:
            lines = notesfile.readlines()
            f.writelines(lines)

    def md_configs(self, filepath: str) -> None:
        """
        Converts the cfg files into markdown at the path specified.
        """
        with open(filepath, mode="w", encoding="utf-8") as f:
            self._md_header(f)
            f.write("* CFG Sections\n")
            for section in self._sections:
                f.write(f"  * [{section}](#{section})\n")
            f.write("* [Report Files](#report_files)\n")
            f.write("* [Notes for Developers](#notes)\n")
            f.write("\n")

            for section in self._sections:
                self._md_section(section, f)
            self.md_reports(f)
            self.md_notes(f)
